import sys
if sys.version_info.major >= 3 and sys.version_info.minor >= 7:  # most code needs at least python 3.7
    from .trainer.trainer_node import TrainerNode
    from .converter.converter_node import ConverterNode

from .detector.detector_node import DetectorNode
from .loop import loop



import os
from typing import List, Optional
from uuid import uuid4
from learning_loop_node.trainer.executor import Executor
from learning_loop_node.trainer.training import Training
from learning_loop_node.trainer.model import BasicModel
from learning_loop_node.context import Context
from learning_loop_node.node import Node
from learning_loop_node.trainer.downloader import TrainingsDownloader
from learning_loop_node.rest import downloads, uploads
from icecream import ic


class Trainer():

    def __init__(self, model_format: str) -> None:
        self.model_format: str = model_format
        self.training: Optional[Training] = None
        self.executor: Optional[Executor] = None

    async def begin_training(self, context: Context, source_model: dict) -> None:
        downloader = TrainingsDownloader(context)

        self.training = Trainer.generate_training(context, source_model)
        self.training.data = await downloader.download_training_data(self.training.images_folder)
        self.executor = Executor(self.training.training_folder)

        await downloads.download_model(self.training.training_folder, context, source_model['id'], self.model_format)

        await self.start_training()

    async def start_training(self) -> None:
        raise NotImplementedError()

    def stop_training(self) -> None:
        raise NotImplementedError()

    def get_error(self) -> str:
        raise NotImplementedError()

    def get_log(self) -> str:
        return self.executor.get_log()

    async def save_model(self,  context: Context, model_id: str) -> None:
        files = self.get_model_files(model_id)
        await uploads.upload_model(context, files, model_id, self.model_format)

    def get_model_files(self, model_id) -> List[str]:
        raise NotImplementedError()

    def get_new_model(self) -> Optional[BasicModel]:
        raise NotImplementedError()

    def on_model_published(self, basic_model: BasicModel, uuid: str) -> None:
        raise NotImplementedError()

    @staticmethod
    def generate_training(context: Context, source_model: dict) -> Training:
        training_uuid = str(uuid4())
        project_folder = Node.create_project_folder(context)
        return Training(
            id=training_uuid,
            base_model=source_model,
            context=context,
            project_folder=project_folder,
            images_folder=Trainer.create_image_folder(project_folder),
            training_folder=Trainer.create_training_folder(project_folder, training_uuid)
        )

    @staticmethod
    def create_image_folder(project_folder: str) -> str:
        image_folder = f'{project_folder}/images'
        os.makedirs(image_folder, exist_ok=True)
        return image_folder

    @staticmethod
    def create_training_folder(project_folder: str, trainings_id: str) -> str:
        training_folder = f'{project_folder}/trainings/{trainings_id}'
        os.makedirs(training_folder, exist_ok=True)
        return training_folder



from typing import List, Optional
import shutil
import os
from glob import glob
import asyncio
from learning_loop_node.context import Context
from learning_loop_node.data_classes.basic_data import BasicData
from learning_loop_node.loop import loop
from learning_loop_node.rest import downloads
from learning_loop_node import node_helper
import logging
from icecream import ic


class DataDownloader():
    context: Context

    def __init__(self, context: Context):
        self.context = context

        self.check_jpeg = shutil.which('jpeginfo') is not None
        if self.check_jpeg:
            logging.info(
                'Detected command line tool "jpeginfo". Images will be checked for validity')
        else:
            logging.error(
                'Missing command line tool "jpeginfo". We can not check for validity of images.')

    async def download_basic_data(self, query_params: Optional[str] = '') -> BasicData:
        async with loop.get(f'api/{self.context.organization}/projects/{self.context.project}/data?{query_params}') as response:
            assert response.status == 200, response
            basic_data = BasicData.parse_obj(await response.json())
            return basic_data

    async def download_images_data(self, ids: List[str]) -> List[dict]:
        return await downloads.download_images_data(self.context.organization, self.context.project, ids)

    async def download_images(self, image_ids: List[str], image_folder: str) -> None:
        paths, ids = node_helper.create_resource_paths(self.context.organization, self.context.project,
                                                       DataDownloader.filter_existing_images(image_ids, image_folder))  # hier
        await downloads.download_images(paths, ids, image_folder)

    async def is_valid_image(self, file):
        if not os.path.isfile(file):
            return False
        if not self.check_jpeg:
            return True

        info = await asyncio.create_subprocess_shell(
            f'jpeginfo -c {file}',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        out, err = await info.communicate()
        return "[OK]" in out.decode()

    @staticmethod
    def filter_existing_images(all_image_ids, image_folder) -> List[str]:
        ids = [os.path.splitext(os.path.basename(image))[0]
               for image in glob(f'{image_folder}/*.jpg')]
        return [id for id in all_image_ids if id not in ids]



from learning_loop_node.context import Context
from typing import List, Optional
from learning_loop_node.node import Node
from learning_loop_node.trainer.trainer import Trainer
import aiohttp
from learning_loop_node import loop
import logging


async def assert_upload_model(file_paths: Optional[List[str]] = None, format: str = 'mocked') -> str:
    if not file_paths:
        file_paths = ['trainer/tests/test_data/file_1.txt',
                      'trainer/tests/test_data/file_2.txt']
    data = [('files', open(path, 'rb')) for path in file_paths]

    data = aiohttp.FormData()

    for path in file_paths:
        data.add_field('files',  open(path, 'rb'))
    async with loop.post(f'api/zauberzeug/projects/pytest/models/{format}', data) as response:
        if response.status != 200:
            msg = f'unexpected status code {response.status} while posting new model'
            logging.error(msg)
            raise(Exception(msg))
        model = await response.json()
        return model['id']


def create_needed_folders(base_folder: str, training_uuid: str = 'some_uuid'):
    project_folder = Node.create_project_folder(
        Context(organization='zauberzeug', project='pytest', base_folder=base_folder))
    image_folder = Trainer.create_image_folder(project_folder)
    training_folder = Trainer.create_training_folder(project_folder, training_uuid)
    return project_folder, image_folder, training_folder



from typing import List, Tuple


def create_resource_paths(organization_name: str, project_name: str, image_ids: List[str]) -> Tuple[List[str], List[str]]:
    if not image_ids:
        return [], []
    url_ids = [(f'api/{organization_name}/projects/{project_name}/images/{id}/main', id)
               for id in image_ids]
    urls, ids = list(map(list, zip(*url_ids)))

    return urls, ids



from pydantic.main import BaseModel
from .context import Context
from typing import List, Optional
from enum import Enum


class CategoryType(str, Enum):
    Box = 'box'
    Point = 'point'
    Segmentation = 'segmentation'


class Category(BaseModel):
    id: str
    name: str
    description: str
    hotkey: str
    color: str
    type: CategoryType


class ModelInformation(BaseModel):
    id: str
    host: str
    organization: str
    project: str
    version: str
    categories: List[Category]
    resolution: Optional[int]

    @property
    def context(self):
        return Context(organization=self.organization, project=self.project)



from typing import Optional
from pydantic import BaseModel
from typing import List


class TrainingData(BaseModel):
    image_data: List[dict]
    categories: List[dict]
    skipped_image_count: Optional[int] = 0

    def image_ids(self):
        return [image['id'] for image in self.image_data]

    def train_image_count(self):
        return len([image for image in self.image_data if image['set'] == 'train'])

    def test_image_count(self):
        return len([image for image in self.image_data if image['set'] == 'test'])



from learning_loop_node.node import Node
from learning_loop_node.status import State


class DetectorNode(Node):

    async def send_status(self):
        # NOTE not yet implemented
        pass

    def get_state(self):
        return State.Running




import psutil
import os
import subprocess
import signal
from icecream import ic
import logging


class Executor:

    def __init__(self, base_path) -> None:
        self.path = base_path
        os.makedirs(self.path, exist_ok=True)
        self.process = None

    def start(self, cmd: str):
        self.process = subprocess.Popen(
            f'cd {self.path}; {cmd} >> last_training.log 2>&1',
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            executable='/bin/bash',
            preexec_fn=os.setsid
        )

    def is_process_running(self):
        if self.process is None:
            return False

        if self.process.poll() is not None:
            return False

        try:
            psutil.Process(self.process.pid)
        except psutil.NoSuchProcess:
            return False

        return True

    def get_log(self):
        with open(f'{self.path}/last_training.log') as f:
            return f.read()

    def stop(self):
        if self.process is None:
            logging.info('no process running ... nothing to stop')
            return

        logging.info('terminating process')

        os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)

        self.process.terminate()
        out, err = self.process.communicate(timeout=3)



from typing import Optional
from pydantic import BaseModel

class SocketResponse(BaseModel):
    success : bool
    error_msg : Optional[str]
    payload: Optional[str]
   
    @staticmethod
    def for_failure(error_msg: str):
        return SocketResponse(success = False, error_msg = error_msg)

    @staticmethod
    def for_success(payload : Optional[str] = ''):
        return SocketResponse(success = True, payload = payload)
    
    


from typing import List
import pytest

import logging
from learning_loop_node.model_information import ModelInformation
from learning_loop_node.converter.converter import Converter
from learning_loop_node.converter.converter_node import ConverterNode


class MockedConverter(Converter):
    models: List[ModelInformation] = []

    async def _convert(self, model_information: ModelInformation) -> None:
        self.models.append(model_information)


@pytest.mark.asyncio
async def test_meta_informations():
    converter = MockedConverter(source_format='mocked', target_format='test')
    node = ConverterNode(name='test', converter=converter)
    await node.convert_models()
    assert len(converter.models) == 3
    categories = converter.models[0].categories
    assert len(categories) == 6
    assert categories[0].type == 'box'
    assert categories[1].type == 'point'



from typing import List
import aiohttp
from learning_loop_node.context import Context
from learning_loop_node.loop import loop
import logging
from icecream import ic


async def upload_model(context: Context, files: List[str], model_id: str, format: str) -> None:
    data = aiohttp.FormData()

    for file_name in files:
        data.add_field('files',  open(file_name, 'rb'))
    async with loop.put(f'api/{context.organization}/projects/{context.project}/models/{model_id}/{format}/file', data=data) as response:
        if response.status != 200:
            msg = f'---- could not save model with id {model_id}. Details: {response.content}'
            raise Exception(msg)
        else:
            logging.info(f'---- uploaded model with id {model_id}')
